from dataclasses import dataclass
import math
import os
import time
import threading

from feathersdk.comms import CanOverloadError, CommsManager, MotorUIDType
from feathersdk.comms.system import S_uint16le
from feathersdk.utils.common import currtime, make_exception_pickleable
from feathersdk.utils.constants import MOTORS_CALIBRATION_PATH
from feathersdk.utils.feathertypes import *
from feathersdk.utils.files import read_json_file, write_json_file
from feathersdk.utils.logger import info, warning
from feathersdk.utils.math_utils import degrees_to_radians

from ..motor import Motor
from .admin import RobstrideAdminControl
from .params import *
from .params import _ROBSTRIDE_PARAMS

if TYPE_CHECKING:  # Due to circular import
    from ..motors_manager import MotorsManager, CalibrationWarnings


T = TypeVar("T")

@dataclass
class TimestampedValue(Generic[T]):
    value: T
    timestamp: float

    def __iter__(self):
        return iter((self.value, self.timestamp))

    def __len__(self):
        return 2

    def __getitem__(self, i: int):
        return (self.value, self.timestamp)[i]
    
    def update(self, value: T, timestamp: Optional[float] = None) -> None:
        """Update the value and timestamp of the TimestampedValue."""
        self.value = value
        self.timestamp = timestamp if timestamp is not None else currtime()


@make_exception_pickleable
class UnsafeCommandError(Exception):
    """Raised when attempting to execute an unsafe command."""
    def __init__(self, message: str, motor: 'RobstrideMotor'):
        super().__init__(message)
        self.motor = motor
        self.motor_name = motor.motor_name


@make_exception_pickleable
class MotorDisabledError(Exception):
    """Raised when attempting to command a motor that is disabled (in Reset mode)."""
    def __init__(self, message: str, motor: 'RobstrideMotor'):
        super().__init__(message)
        self.motor = motor
        self.motor_name = motor.motor_name


@make_exception_pickleable
class MotorModeInconsistentError(Exception):
    """Raised when attempting to command a motor with an incompatible run mode."""
    def __init__(self, message: str, motor: 'RobstrideMotor', curr_mode: RobstrideRunMode, exp_mode: RobstrideRunMode):
        super().__init__(message)
        self.motor = motor
        self.motor_name = motor.motor_name
        self.current_mode = curr_mode
        self.expected_mode = exp_mode


_MOTOR_NO_DEFAULT = object()


class RobstrideMotorConfig(TypedDict, total=False):
    version: RobstrideVersionType
    max_torque: float  # Max torque in N
    max_position_dx: float  # positional difference in radians
    joint_limits: Tuple[float, float]  # joint limits in radians
    max_velocity: float  # Max velocity in radians / seconds


class RobstrideMotor(Motor):
    """Class to control a Robstride motor.
    
    Args:
        motor_id: The motor ID to use for the communication.
        name: The name of the motor.
        config: The configuration for the motor.
        iface: The interface to use for the communication.
    """
    # Class-level list of feedback listeners - empty list is falsy for fast check
    _feedback_listeners: List[Callable[['RobstrideMotor', float], None]] = []

    # Class-level list of command listeners - notified when loc_ref or spd_ref is written
    _command_listeners: List[Callable[['RobstrideMotor', str, float, float], None]] = []

    @classmethod
    def add_feedback_listener(cls, listener: Callable[['RobstrideMotor', float], None]) -> None:
        """Register a callback to receive feedback events after each update_feedback call."""
        cls._feedback_listeners.append(listener)

    @classmethod
    def remove_feedback_listener(cls, listener: Callable[['RobstrideMotor', float], None]) -> None:
        """Unregister a feedback listener."""
        cls._feedback_listeners.remove(listener)

    @classmethod
    def add_command_listener(
        cls, listener: Callable[['RobstrideMotor', str, float, float], None]
    ) -> None:
        """Register a callback for target position/velocity commands (loc_ref, spd_ref)."""
        cls._command_listeners.append(listener)

    @classmethod
    def remove_command_listener(
        cls, listener: Callable[['RobstrideMotor', str, float, float], None]
    ) -> None:
        """Unregister a command listener."""
        cls._command_listeners.remove(listener)

    def __init__(self, motor_id: int, name: str, config: RobstrideMotorConfig = {}, iface: Optional[str] = None):
        if "model" in config and "version" not in config:
            config["version"] = config["model"]
        if "version" not in config:
            raise KeyError("\"version\" key is required for RobstrideMotorConfig")
        self.version = get_robstride_version(config["version"])

        if motor_id < 0 or motor_id >= 0xFE:
            raise ValueError(f"Error: motor_id must be between [0, 0xFD], got {motor_id}")

        self.manager: Optional['MotorsManager'] = None

        self.comms = CommsManager()
        self.motor_id: int = motor_id
        self.default_host_id: int = 0xAA
        self.uid: MotorUIDType = MotorUIDType(self.motor_id)
        self.motor_name: str = name
        self.calibration_time = None
        self._homing_pos = config.get("homing_pos", 0)
        self.direction = config.get("direction", 1)
        self.enable_lock = threading.Lock()

        self.range = config.get("range", {
            "min": -math.pi / 2,
            "max": math.pi / 2,
        })

        self.upper_limit = None
        self.lower_limit = None
        # calibration_homing_pos stores the actual encoder value where the motor's 
        # homing position is located after calibration. This may have a small residual 
        # offset from 0 due to motor drift after zeroing.
        # This value is used as an internal offset for motor commands and calibrated_angle.
        self.calibration_homing_pos = None
        self.total_range = None
        self.last_kp = config.get("starting_kp", 30)
        self.last_kd = config.get("starting_kd", 5)
        self._should_monitor = False
        
        self.calibrated_angle = (0, -1)
        self.dual_encoder: bool = self._get_dual_encoder_property()

        self.family_name: Union[str, None] = None
        self.iface: Union[str, None] = iface
        self.properties: Dict[str, TimestampedValue] = {}
        self.joint_limits = config.get("joint_limits", None)

        if self.joint_limits is not None:
            if (
                self.joint_limits[0] < -math.pi
                or self.joint_limits[1] > math.pi
                and self.joint_limits[0] <= self.joint_limits[1]
            ):
                print("Warning: Joint limits must be from -pi to pi for ", name)
        self.max_torque = config.get("max_torque", None)
        self.max_position_dx = config.get("max_position_dx", None)
        self.max_velocity = config.get("max_velocity", None)
        self.run_mode = RobstrideRunMode.Operation
        self.compliance_mode: bool = False
        self._disable_compliance_mode: bool = False
        # Target position is the target angle to reach. It's relative to angle.
        self.target_position = 0
        self.last_compliance_pos = 0
        self.compliance_mode_torque_threshold = config.get("compliance_mode_torque_threshold", 1.0)
        self.compliance_mode_dx = config.get("compliance_mode_dx", 0.01)

        self.mode: TimestampedValue[RobstrideMotorMode] = TimestampedValue(RobstrideMotorMode.Reset, -1)
        # Angle is the raw angle from the encoder.
        self.angle: TimestampedValue[float] = TimestampedValue(0, -1)
        # Calibrated angle is around 0 that should be used for URDFs. It is the angle - calibration_homing_pos.
        self.calibrated_angle: TimestampedValue[float] = TimestampedValue(0, -1)
        self.velocity: TimestampedValue[float] = TimestampedValue(0, -1)
        self.torque: TimestampedValue[float] = TimestampedValue(0, -1)
        self.temp: TimestampedValue[float] = TimestampedValue(0, -1)
        self.errors: TimestampedValue[List[RobstrideMotorError]] = TimestampedValue([], -1)

        # Feedback delay detection
        self._last_command_time: float = 0.0  # Time when last feedback-triggering command was sent
        self._feedback_delay_threshold: float = config.get("feedback_delay_threshold", 1.0)
        self._feedback_delay_detected: bool = False

        self.load_calibration_state()

        self._admin: RobstrideAdminControl = RobstrideAdminControl(self)

    def set_monitor(self, should_monitor: bool = True) -> None:
        """Set whether to monitor the motor."""
        self._should_monitor = should_monitor

    def is_runable(self) -> bool:
        """ Returns True if the motor is a state where it's ready to listen to and execute motion commands. """
        return self.mode[0] == RobstrideMotorMode.Run and not self.compliance_mode

    def _get_dual_encoder_property(self) -> bool:
        if self.version == RobstrideVersion.RS01:
            return False
        elif self.version in [RobstrideVersion.RS00, RobstrideVersion.RS02, RobstrideVersion.RS03,
                              RobstrideVersion.RS04, RobstrideVersion.RS05, RobstrideVersion.RS06]:
            return True
        else:
            raise NotImplementedError(f"Unknown motor version for dual_encoder logic: {self.version}")

    @property
    def expected_range(self) -> float:
        return self.range["max"] - self.range["min"]

    def _record_command_sent(self) -> None:
        """Record that a command expecting feedback was just sent."""
        self._last_command_time = currtime()

    def _log_feedback_delay(self, delay_seconds: float) -> None:
        """Log warning message when feedback delay is detected."""
        warning(
            f"FEEDBACK DELAY: Motor '{self.motor_name}' (ID: {self.motor_id}) "
            f"delayed {delay_seconds:.3f}s (threshold: {self._feedback_delay_threshold}s)"
        )

    def _log_robstride_error_changes(
        self, new_errors: Iterable[RobstrideMotorError], source: str
    ) -> None:
        """Log when drive-reported fault bits appear or clear (compared to previous stored errors)."""
        if not new_errors and not self.errors[0]:
            return

        prior = set(self.errors[0]) if self.errors[0] else set()
        new_set = set(new_errors) if new_errors else set()

        if prior == new_set:
            return

        added = new_set - prior
        removed = prior - new_set
        if added:
            names = ", ".join(e.name for e in sorted(added, key=lambda e: e.value))
            warning(
                f"Robstride motor '{self.motor_name}' (id=0x{self.motor_id:02x}) [{source}]: "
                f"fault(s) active: {names}"
            )
        if removed:
            names = ", ".join(e.name for e in sorted(removed, key=lambda e: e.value))
            warning(
                f"Robstride motor '{self.motor_name}' (id=0x{self.motor_id:02x}) [{source}]: "
                f"fault(s) cleared: {names}"
            )

    @property
    def feedback_delay_detected(self) -> bool:
        """Returns True if feedback delay was detected on the last command."""
        return self._feedback_delay_detected

    def get_pending_feedback_time(self) -> float:
        """Returns time since last command awaiting feedback, or 0 if none pending."""
        if self._last_command_time <= 0:
            return 0.0
        return currtime() - self._last_command_time

    def check_feedback_timeout(self, threshold: Optional[float] = None) -> Optional[float]:
        """Check if waiting for feedback that hasn't arrived within threshold.
        
        Args:
            threshold: Optional timeout threshold in seconds. If not provided,
                uses the motor's configured _feedback_delay_threshold.
        
        Returns:
            Delay in seconds if timeout detected, None otherwise.
        """
        if self._last_command_time <= 0:
            return None
        elapsed = currtime() - self._last_command_time
        effective_threshold = threshold if threshold is not None else self._feedback_delay_threshold
        if elapsed > effective_threshold:
            self._log_feedback_delay(elapsed)
            self._feedback_delay_detected = True
            return elapsed
        return None

    @property
    def homing_pos(self) -> float:
        """Get the encoder value for the homing position.
        
        Returns the calibration_homing_pos plus any config-defined offset (_homing_pos).
        After calibrate_motor is called, calibration_homing_pos contains the actual 
        encoder value at homing, which may have a small residual offset from 0.
        
        Returns:
            The encoder value to command when moving to homing position
        """
        if self.calibration_homing_pos is not None:
            return self.calibration_homing_pos + self._homing_pos
        return self._homing_pos

    def update_angle(self, angle: float):
        self.angle = TimestampedValue(angle, currtime())
        if self.calibration_homing_pos is not None:
            self.calibrated_angle = TimestampedValue(self.convert_angle_to_calibrated_angle(angle), currtime())
        else:
            self.calibrated_angle = TimestampedValue(angle, currtime())

    def convert_calibrated_angle_to_angle(self, calibrated_angle: float) -> float:
        if self.calibration_homing_pos is None:
            return calibrated_angle
        return calibrated_angle + self.calibration_homing_pos

    def convert_angle_to_calibrated_angle(self, angle: float) -> float:
        if self.calibration_homing_pos is None:
            return angle
        return angle - self.calibration_homing_pos

    def update_feedback(
        self, angle: float, velocity: float, torque: float, temp: float, errors: List[RobstrideMotorError], mode: RobstrideMotorMode
    ):
        last_update = currtime()
        
        # Check for feedback delay
        if self._last_command_time > 0:
            feedback_delay = last_update - self._last_command_time
            if feedback_delay > self._feedback_delay_threshold:
                self._log_feedback_delay(feedback_delay)
                self._feedback_delay_detected = True
            else:
                self._feedback_delay_detected = False
            self._last_command_time = 0.0  # Clear pending command
        
        # Raw angle coming from the encoder.
        self.update_angle(angle)
        self.velocity = TimestampedValue(velocity, last_update)
        self.torque = TimestampedValue(torque, last_update)
        self.temp = TimestampedValue(temp, last_update)
        self._log_robstride_error_changes(set(errors) if errors else set(), "motor_feedback")
        self.errors = TimestampedValue(errors, last_update)
        self.mode = TimestampedValue(mode, last_update)

        # Event notification - empty list check is O(1) and falsy
        if RobstrideMotor._feedback_listeners:
            for listener in RobstrideMotor._feedback_listeners:
                listener(self, last_update)

    def update_property(self, property_name: str, value: Union[float, int]):
        self.properties[property_name] = TimestampedValue(value, currtime())
    
    def get_property_value(self, property_name: str, default: Any = _MOTOR_NO_DEFAULT) -> Any:
        """Get the currently stored value for a property in this motor. 
        
        If the property is not found, return the default value. If the default value is not provided, raise a KeyError.
        """
        if property_name in self.properties:
            return self.properties[property_name].value
        if default is not _MOTOR_NO_DEFAULT:
            return default
        raise KeyError(f"Property {property_name} not found in motor {self.motor_name}")

    def get_property_timestamp(self, property_name: str, default: Any = _MOTOR_NO_DEFAULT) -> Any:
        """Get the timestamp of the currently stored value for a property in this motor. 
        
        If the property is not found, return the default value. If the default value is not provided, raise a KeyError.
        """
        if property_name in self.properties:
            return self.properties[property_name].timestamp
        if default is not _MOTOR_NO_DEFAULT:
            return default
        raise KeyError(f"Property {property_name} not found in motor {self.motor_name}")

    def is_safe_position_update(self, target_position: float) -> bool:
        if self.max_position_dx is not None:
            return abs(target_position - self.angle[0]) <= self.max_position_dx
            # if not abs(target_position - self.angle[0]) <= self.max_position_dx:
            #     return abs(target_position - self.target_position) <= self.max_position_dx
            # else:
            #     return True
        return True

    def update_target_position(self, target_position: float, from_compliance_mode: bool = False) -> None:
        """
        Used for safety purposes.
        If max_position_dx is set, the motor will be stopped if the position is outside the max_position_dx.
        """
        if self.compliance_mode and not from_compliance_mode:
            raise UnsafeCommandError(message="Commands currently blocked - in compliance mode", motor=self)

        self.target_position = target_position

    def should_trigger_compliance_mode(self):
        if self._disable_compliance_mode:
            return False

        if self.compliance_mode or self.mode[0] == RobstrideMotorMode.Reset:
            return False

        if self.max_position_dx is not None:
            if abs(self.angle[0] - self.target_position) > self.max_position_dx:
                warning(
                    f"Compliance mode triggered: Motor {self.motor_name} Position {self.angle[0]} is outside max position dx {self.max_position_dx} Target position {self.target_position}"
                )
                return True

        if self.max_torque is not None:
            if abs(self.torque[0]) > self.max_torque:
                warning(
                    f"Compliance mode triggered: Motor {self.motor_name} Torque {self.torque[0]} is greater than max torque {self.max_torque}"
                )
                return True

        if self.joint_limits is not None:
            if self.calibrated_angle[0] > self.joint_limits[1] or self.calibrated_angle[0] < self.joint_limits[0]:
                warning(
                    f"Compliance mode triggered: Motor {self.motor_name} Angle {self.calibrated_angle[0]} is outside joint limits {self.joint_limits}"
                )
                return True

        if self.max_velocity is not None:
            if abs(self.velocity[0]) > self.max_velocity:
                warning(
                    f"Compliance mode triggered: Motor {self.motor_name} Velocity {self.velocity[0]} is greater than max velocity {self.max_velocity}"
                )
                return True

        return False

    def _disable_compliance_mode_for_calibration(self):
        """Disable compliance mode triggering during calibration."""
        self._disable_compliance_mode = True

    def _enable_compliance_mode_after_calibration(self):
        """Re-enable compliance mode triggering after calibration."""
        self._disable_compliance_mode = False

    def save_calibration_state(self, lower_limit: float, upper_limit: float, calibration_homing_pos: float, total_range: float):
        """Set the calibration data for this motor.
        
        Args:
            lower_limit: The encoder value at the lower physical limit
            upper_limit: The encoder value at the upper physical limit
            calibration_homing_pos: The encoder value at the homing position after calibration.
                This may have a small residual offset from 0 due to motor drift after zeroing.
            total_range: The total range of motion in radians
        """
        self._disable_compliance_mode = False
        # we must use time.time() here instead of monotonic because
        # we need to compare this with system power up time which is more inline with time.time().
        self.calibration_time = time.time()
        self.lower_limit = lower_limit
        self.upper_limit = upper_limit
        self.calibration_homing_pos = calibration_homing_pos
        self.total_range = total_range

        data = {
            "calibration_time": self.calibration_time,
            "lower_limit": self.lower_limit,
            "upper_limit": self.upper_limit,
            "calibration_homing_pos": self.calibration_homing_pos,
            "total_range": self.total_range,
        }
        write_json_file(data, MOTORS_CALIBRATION_PATH + "/" + self.motor_name + ".json")

    def load_calibration_state(self) -> bool:
        """Load the calibration state from the file.
        
        Returns True if the calibration state was loaded successfully, False otherwise.
        Returns False if the file doesn't exist or uses the old calibration format
        (middle_pos instead of calibration_homing_pos), requiring recalibration.
        """
        file_path = MOTORS_CALIBRATION_PATH + "/" + self.motor_name + ".json"
        if not os.path.exists(file_path):
            return False
        data = read_json_file(file_path)
        # Check for new calibration format - return False if using old format
        if "calibration_homing_pos" not in data:
            return False
        self.calibration_time = data["calibration_time"]
        self.lower_limit = data["lower_limit"]
        self.upper_limit = data["upper_limit"]
        self.calibration_homing_pos = data["calibration_homing_pos"]
        self.total_range = data["total_range"]
        return True

    def _calibration_checks(self) -> 'CalibrationWarnings':
        """Check calibration values and return any warnings.
        
        This method should be called after a motor is calibrated to validate
        that the calibration values are within expected tolerances.
        
        Returns:
            CalibrationWarnings flags indicating any issues found
        """
        # Import here to avoid circular import at module level
        from ..motors_manager import CalibrationWarnings
        
        assert self.calibration_time is not None, "This method should be called after a motor is calibrated."
        warnings = CalibrationWarnings.NONE
        if abs(self.expected_range - self.total_range) > degrees_to_radians(1.0):
            warnings |= CalibrationWarnings.RANGE_MISMATCH
        if abs(self.calibration_homing_pos) > degrees_to_radians(5.0):
            warnings |= CalibrationWarnings.HOMING_POSITION_NONZERO
        three_degrees_in_radians = degrees_to_radians(3.0)
        if abs(self.lower_limit - self.range["min"]) > three_degrees_in_radians:
            warnings |= CalibrationWarnings.LOWER_LIMIT_MISMATCH
        if abs(self.upper_limit - self.range["max"]) > three_degrees_in_radians:
            warnings |= CalibrationWarnings.UPPER_LIMIT_MISMATCH
        return warnings

    def _msg_can_id(self, op: RobstrideMotorMsg, motor_id: int, host_id: Union[int, bytes, None] = None, e_byte: int = 0) -> int:
        """Build a CAN ID for a Robstride motor message.
        
        Args:
            op: The operation to build the CAN ID for.
            motor_id: The motor ID to build the CAN ID for.
            host_id: The host ID to build the CAN ID for. If int, then data will be checked to be within valid bounds.
                If you wish to skip this check, pass it as a bytes object with one element.
            e_byte: The extra byte to build the CAN ID for.
        """
        if isinstance(host_id, bytes):
            assert len(host_id) == 1, f"Error: host_id must be 1 byte, got {len(host_id)}"
            host_id = host_id[0]
        else:
            host_id = self.default_host_id if host_id is None else host_id
            if host_id < 0 or host_id > 0xFD:
                raise ValueError(f"Error: host_id must be between [0, 0xFD], got {host_id}")
        
        if motor_id < 0 or motor_id > 0xFD:
            raise ValueError(f"Error: motor_id must be between [0, 0xFD], got {motor_id}")
        if e_byte < 0 or e_byte > 0xFF:
            raise ValueError(f"Error: e_byte must be between [0, 0xFF], got {e_byte}")
        
        return (op.value << 24) | (e_byte << 16) | (host_id << 8) | motor_id
    
    def get_device_id(self) -> int:
        """Get the device id of the motor."""
        can_id = self._msg_can_id(RobstrideMotorMsg.GetDeviceId, self.motor_id)
        self.comms.cansend(self.iface, True, can_id, bytes([0] * 8), self.uid)

    def enable(self):
        """Send the 'enable' command to the motor."""
        can_id = self._msg_can_id(RobstrideMotorMsg.Enable, self.motor_id)
        self.comms.cansend(self.iface, True, can_id, bytes([0] * 8), self.uid)
        self._record_command_sent()

    def disable(self, clear_faults: bool = False):
        """Send the 'disable' command to the motor.
        
        Args:
            clear_faults: If True, clear the fault bits in the disable command.
        """
        can_id = self._msg_can_id(RobstrideMotorMsg.Disable, self.motor_id)
        self.comms.cansend(self.iface, True, can_id, bytes([1] + [0] * 7) if clear_faults else bytes([0] * 8), self.uid)
        self._record_command_sent()
    
    def set_active_reporting(self, active: bool, scan_time_ms: int = 10):
        """Turn on/off active reporting for the motor.
        
        Args:
            active: If True, turn on active reporting. False to disable.
            scan_time_ms: Decides the time between feedback messages. Minimum is 10ms, max is 327680ms (about 5.5
                minutes). Note: The motor only supports multiples of 5ms, so the actual scan time will be rounded
                down to the nearest multiple of 5ms.
        """
        if scan_time_ms < 10 or scan_time_ms > 327680:
            raise ValueError(f"Invalid scan time: {scan_time_ms}ms. Must be between 10ms and 327680ms.")
        
        # For the motor, 1 = 10ms, +1 = +5ms
        self.write_param(_ROBSTRIDE_PARAMS.epscan_time, int(scan_time_ms) // 5 - 1)

        can_id = self._msg_can_id(RobstrideMotorMsg.SetMotorActiveReporting, self.motor_id)
        data = bytes([1, 2, 3, 4, 5, 6, 1 if active else 0, 8])
        self.comms.cansend(self.iface, True, can_id, data, self.uid)
    
    def zero_position(self):
        """Set the zero position for the motor."""
        can_id = self._msg_can_id(RobstrideMotorMsg.ZeroPos, self.motor_id)
        self.comms.cansend(self.iface, True, can_id, bytes([0x01] + [0] * 7), self.uid)
        self._record_command_sent()

    def read_param(self, param_id: Union[int, str, RobstrideParam]) -> float:
        """Read a parameter from the motor."""
        can_id = self._msg_can_id(RobstrideMotorMsg.ReadParam, self.motor_id)
        data = S_uint16le.pack(get_param_info(param_id).id) + bytes([0] * 6)
        self.comms.cansend(self.iface, True, can_id, data, self.uid)

    def write_param(self, p_id: Union[int, str, RobstrideParam], value: Union[float, int]) -> None:
        self._write_param_checked(p_id, value, force_run=False)

    def _write_param_checked(self, p_id: Union[int, str, RobstrideParam], value: Union[float, int], force_run: bool = False) -> None:
        """Write a parameter value to the specified motor.

        Args:
            p_id: The parameter ID to write.
            value: The value to write.
            force_run: If True, force the write to be executed, ignoring the safety checks.
        """
        param = get_param_info(p_id)
        if not force_run and not param.is_within_range(value, self.version):
            raise ValueError(f"Value {value} is out of range for param {param.name}: {param.range[self.version]}")
            
        data = S_uint16le.pack(param.id) + bytes([0] * 2) + param.dtype.pack(value, pad_end=4, pad_val=0)
        
        if not force_run and param.name == "loc_ref":
            if self.is_safe_position_update(value):
                self.update_target_position(value)  # can raise an exception to cancel the run.
            else:
                warning(f"Triggering compliance mode [{self.motor_name}]: Target position {value} is too far from current position {self.angle[0]} last target position {self.target_position} (max dx: {self.max_position_dx})")
                self.trigger_compliance_mode()
                raise UnsafeCommandError(
                    message=f"Safety Error: Target position {value} is too far from current position {self.angle[0]} last target position {self.target_position} (max dx: {self.max_position_dx})",
                    motor=self
                ) # stop the program safely.

        can_id = self._msg_can_id(RobstrideMotorMsg.WriteParam, self.motor_id)

        try:
            self.comms.cansend(self.iface, True, can_id, data, self.uid)
            self._record_command_sent()
        except CanOverloadError:
            time.sleep(0.01)
            self.comms.cansend(self.iface, True, can_id, data, self.uid)
            self._record_command_sent()

        if param.name in ("loc_ref", "spd_ref") and RobstrideMotor._command_listeners:
            ts = currtime()
            for listener in RobstrideMotor._command_listeners:
                listener(self, param.name, value, ts)
        
    def set_run_mode(self, run_mode: RobstrideRunMode):
        self.write_param("run_mode", run_mode.value)
        self.run_mode = run_mode

    def set_target_position(self, target_position: float) -> None:
        """Set the target position (loc_ref) for the specified motor."""
        # Check if motor is disabled (in Reset mode)
        if self.mode[0] == RobstrideMotorMode.Reset:
            raise MotorDisabledError(
                message=f"Cannot set target position for motor {self.motor_name}: "
                        f"Motor is disabled (in Reset mode). Enable the motor first.",
                motor=self
            )
        
        # Check if motor is in the correct run mode for position control
        if self.run_mode != RobstrideRunMode.Position:
            raise MotorModeInconsistentError(
                f"Cannot set target position for motor {self.motor_name}: "
                f"Motor is in {self.run_mode.name} mode, but Position mode is required. "
                f"Use set_run_mode() to switch to Position mode first.",
                motor=self,
                curr_mode=self.run_mode,
                exp_mode=RobstrideRunMode.Position
            )
        
        self.write_param("loc_ref", target_position)
    
    def set_target_velocity(self, target_velocity: float) -> None:
        """Set the target velocity (spd_ref) for the specified motor."""
        # Check if motor is disabled (in Reset mode)
        if self.mode[0] == RobstrideMotorMode.Reset:
            raise MotorDisabledError(
                message=f"Cannot set target velocity for motor {self.motor_name}: "
                        f"Motor is disabled (in Reset mode). Enable the motor first.",
                motor=self
            )
        
        # Check if motor is in the correct run mode for velocity control
        if self.run_mode != RobstrideRunMode.Speed:
            raise MotorModeInconsistentError(
                f"Cannot set target velocity for motor {self.motor_name}: "
                f"Motor is in {self.run_mode.name} mode, but Speed mode is required. "
                f"Use set_run_mode() to switch to Speed mode first.",
                motor=self,
                curr_mode=self.run_mode,
                exp_mode=RobstrideRunMode.Speed
            )
        
        self.write_param("spd_ref", target_velocity)
    
    def operation_command(self, op: RobstrideOperationCommand, apply_target_position: bool = True, from_compliance_mode: bool = False):
        """ target_torque range (-60Nm to 60Nm)"""
        motor = op.motor
        motor.last_kp = op.kp
        motor.last_kd = op.kd        
        if self._should_monitor:
            print(self.motor_name, op.target_angle, op.target_torque, op.target_velocity, motor.last_kp, motor.last_kd, motor.torque)
        can_id, data = op.build_command()

        if apply_target_position:
            if self.is_safe_position_update(op.target_angle) or from_compliance_mode:
                self.update_target_position(op.target_angle, from_compliance_mode)  # can raise an exception to cancel the run.
            else:
                warning(f"Triggering compliance mode [{self.motor_name}]: Target position {op.target_angle} is too far from current position {self.angle[0]} last target position {self.target_position} (max dx: {self.max_position_dx})")
                self.trigger_compliance_mode()
                raise UnsafeCommandError(
                    message=f"Safety Error: Target position {op.target_angle} is too far from current position {self.angle[0]} last target position {self.target_position} (max dx: {self.max_position_dx})",
                    motor=self
                ) # stop the program safely.

        self.comms.cansend(self.iface, True, can_id, data, self.uid)
        self._record_command_sent()
    
    def trigger_compliance_mode(self):
        self.manager.trigger_compliance_mode(self.family_name, self.motor_name)


MotorMap = Dict[str, RobstrideMotor]




#!/usr/bin/env python3
import canopen
import time
import argparse
from enum import Enum

# --- Constants & Enums (Following torso_platform.py style) ---

class EZMotionCommands(Enum):
    """Object Dictionary mapping for EZMotion C2 Series."""
    CONTROL_WORD = 0x6040
    INTERNAL_STATUS = 0x200B  # Sub-indices: 0x09 Max V, 0x0A Min V, 0x0B Faults
    NODE_ID_SETTING = 0x2101
    STORE_PARAMETERS = 0x200D

class EZMotionValues:
    """Special hex codes for specific commands."""
    SAVE_TO_NVM = 0x65766173  # "save" in hex
    FAULT_RESET = 0x80        # Bit 7 high
    
# Configuration Defaults
DEFAULT_EDS_PATH = "./MMS760400-48-C2-1.eds"
BITRATE = 1000000
VOLTAGE_LOWER_LIMIT = 52
VOLTAGE_UPPER_LIMIT = 65

class EZMotionConfigurator:
    def __init__(self, channel='can3', bitrate=BITRATE, eds_path=DEFAULT_EDS_PATH):
        self.network = canopen.Network()
        self.channel = channel
        self.bitrate = bitrate
        self.eds_path = eds_path

    def connect(self):
        print(f"Connecting to {self.channel} at {self.bitrate}bps...")
        self.network.connect(channel=self.channel, bustype="socketcan", bitrate=self.bitrate)

    def disconnect(self):
        self.network.disconnect()
        print("Network disconnected.")

    def update_node_id(self, current_id, new_id):
        try:
            # 1. Add the node with the current ID
            node = self.network.add_node(current_id, self.eds_path)
            print(f"Successfully added Node {current_id}")

            # 2. Read current diagnostics
            fault_register = node.sdo[EZMotionCommands.INTERNAL_STATUS.value][0x0B].raw
            current_node_id = node.sdo[EZMotionCommands.NODE_ID_SETTING.value].raw
            max_v = node.sdo[EZMotionCommands.INTERNAL_STATUS.value][0x09].raw
            min_v = node.sdo[EZMotionCommands.INTERNAL_STATUS.value][0x0A].raw

            print(f"--- Diagnostic Report (Node {current_id}) ---")
            print(f"Fault Register: {hex(fault_register)}")
            print(f"Current Configured ID: {current_node_id}")
            print(f"Voltage Protection: {min_v}V (Min) - {max_v}V (Max)")

            # 3. Update Voltage Protection Limits
            print(f"\nUpdating voltage limits to {VOLTAGE_LOWER_LIMIT}V - {VOLTAGE_UPPER_LIMIT}V...")
            node.sdo[EZMotionCommands.INTERNAL_STATUS.value][0x0A].raw = VOLTAGE_LOWER_LIMIT
            node.sdo[EZMotionCommands.INTERNAL_STATUS.value][0x09].raw = VOLTAGE_UPPER_LIMIT

            # 4. Change the Node ID
            print(f"Changing Node ID from {current_id} to {new_id}...")
            node.sdo[EZMotionCommands.NODE_ID_SETTING.value].raw = new_id

            # 5. Save to NVM (Non-Volatile Memory)
            # This is critical so the ID doesn't revert to default on power cycle
            print("Saving changes to NVM...")
            node.sdo[EZMotionCommands.STORE_PARAMETERS.value].raw = EZMotionValues.SAVE_TO_NVM
            
            # 6. Reset Faults (Optional, but included from your original logic)
            # node.sdo[EZMotionCommands.CONTROL_WORD.value].raw = EZMotionValues.FAULT_RESET
            
            # Short sleep to allow the motor to finish the NVM write
            time.sleep(0.5)
            print(f"Success. Motor will respond as ID {new_id} after next power cycle or reset.")

        except Exception as e:
            print(f"Error communicating with Node {current_id}: {e}")

def main():
    parser = argparse.ArgumentParser(description="EZ Motion Motor ID and Voltage Configurator")
    parser.add_argument("--current_id", type=int, default=2, help="The current CAN ID of the motor (default: 2)")
    parser.add_argument("--new_id", type=int, default=9, required=True, help="The new CAN ID to assign to the motor")
    parser.add_argument("--channel", type=str, default="can0", help="CAN interface (default: can0)")
    
    args = parser.parse_args()

    configurator = EZMotionConfigurator(channel=args.channel)
    
    try:
        configurator.connect()
        configurator.update_node_id(args.current_id, args.new_id)
    finally:
        configurator.disconnect()

if __name__ == "__main__":
    main()



from setuptools import setup

# All package metadata/configuration lives in `pyproject.toml` (PEP 621 + tool.setuptools).
# This file is kept only for compatibility with legacy tooling.
if __name__ == "__main__":
    setup()



"""Misc helper functions that can be used throughout the project"""
import functools
import types
import inspect
import threading
import traceback
import pprint
from inspect import signature, Parameter
from feathersdk.utils.feathertypes import *
from enum import Enum
import time
import os
import re
from feathersdk.utils.logger import warning, error


TExc = TypeVar('TExc', bound=Exception)
TEnum = TypeVar('TEnum', bound=Enum)
T = TypeVar('T')
P = ParamSpec('P')
R = TypeVar('R')


try:
    import tqdm
    def progressbar(iterable: Iterable[T], progress: bool = True) -> Iterable[T]:
        return tqdm.tqdm(iterable, desc="Progress", disable=not progress)
except ImportError:
    def progressbar(iterable: Iterable[T], progress: bool = True) -> Iterable[T]:
        return iterable


def annotation_allows_type(annotation: Any, type_to_check: type) -> bool:
    """Check if a type annotation allows for the given type."""
    if isinstance(annotation, str):
        annotation = eval(annotation)
    
    if annotation in [Parameter.empty, Any, type_to_check]:
        return True
    
    # Recursively check if any of the union args allow the given type
    if get_type_origin(annotation) in [Union, getattr(types, 'UnionType', None)]:
        return any(annotation_allows_type(arg, type_to_check) for arg in get_type_args(annotation))
    
    return False


def make_exception_pickleable(cls: Type[TExc]) -> Type[TExc]:
    """Decorate an exception class to modify it to be pickleable.

    This is useful for exceptions that are raised in multiprocessing Pool calls as they get pickled and sent to the
    main process to be re-raised. During the unpickling process, the __init__ method is called and passed a single 
    string argument: the full exception message from the original pickled exception.This means if you create a new 
    exception class which doesn't take any arguments, or takes arguments which are used to build the exception message, 
    you'll get an error when unpickling due to incorrect number of arguments.

    This function modifies the __init__ method. When called, the first argument is expected to be `self`. For most
    calls, the class-overridden __init__ method will be called with the passed args/kwargs. However, if the new
    __init__ method is called with exactly one argument which is positional and a string, then the superclass's
    __init__ method will be called with the string argument, which is expected to be the full final exception message.

    NOTE: If there is ambiguity in the arguments to the overridden __init__ method (meaning: it is at all possible to 
    pass a single string parameter and the function call would be valid), we may not be able to tell if the argument is
    the exception message or the normal positional argument, and thus an exception will be raised. In this case, you can
    either set a non-string type annotation on what you pass in to the __init__ method, or make the argument 
    keyword-only.

    Args:
        cls: The exception class to modify. Should inherit from Exception.
    """
    old_init = cls.__init__

    # Inspect the different parameters in the __init__ method
    varpos_arg, pos_args, def_args = None, [], []
    for p in signature(old_init).parameters.values():
        if p.name == 'self':
            continue
        elif p.kind == Parameter.VAR_POSITIONAL:
            varpos_arg = p
        elif p.kind in [Parameter.POSITIONAL_ONLY, Parameter.POSITIONAL_OR_KEYWORD]:
            if p.default is Parameter.empty:
                pos_args.append(p)
            else:
                def_args.append(p)
    
    # Check args for ambiguity
    if len(pos_args) <= 1:
        possible_def = [def_args[0]] if len(def_args) > 0 and len(pos_args) == 0 else []
        possible_varpos = [varpos_arg] if varpos_arg is not None and len(pos_args) == 0 else []
        params = possible_varpos + pos_args + possible_def
        if any(annotation_allows_type(p.annotation, str) for p in params):
            raise TypeError(f"Class \"{cls.__name__}\" has ambiguous function signature. Set type annotations to "
                            f"non-string type, or make the arguments keyword-only to avoid this error.")

    def new_init(self: Exception, *args: Any, **kwargs: Any) -> None:
        """Call super's init if we have exactly one arg, no kwargs, and the arg is a string.
        Otherwise call the original __init__ method with the arguments."""
        if len(args) == 1 and len(kwargs) == 0 and isinstance(args[0], str):
            return super(cls, self).__init__(*args)
        return old_init(self, *args, **kwargs)
    
    cls.__init__ = new_init
    return cls


_DEBUG_LOCK_PRINT = False
def set_debug_lock_print(debug_lock_print: bool) -> None:
    """Set whether to print debug messages when acquiring and releasing locks."""
    global _DEBUG_LOCK_PRINT
    _DEBUG_LOCK_PRINT = debug_lock_print


class LockAcquisitionError(Exception):
    """Raised when a lock cannot be acquired."""


class FeatherThreadLock:
    """A custom threading.Lock. Currently can be used to debug print messages. Call set_debug_lock_print() to enable
    
    Parameters
    ----------
    name: ``str``
        The name of the lock.
    timeout: ``float``
        The timeout in seconds to wait for the lock to be acquired.
    pre_acquire_callback: ``Optional[Callable[[], None]]``
        A callback to be called before the lock is acquired.
    post_release_callback: ``Optional[Callable[[], None]]``
        A callback to be called after the lock is released.
    """
    def __init__(self, name: str, timeout: float = 5.0, pre_acquire_callback: Optional[Callable[[], None]] = None, 
                 post_release_callback: Optional[Callable[[], None]] = None):
        self.lock = threading.Lock()
        self.name = name
        self.timeout = timeout
        self.pre_acquire_callback = pre_acquire_callback
        self.post_release_callback = post_release_callback
    
    def is_locked(self) -> bool:
        """Check if the lock is currently locked."""
        return self.lock.locked()
    
    def __enter__(self):
        if _DEBUG_LOCK_PRINT:
            self.func_name = inspect.currentframe().f_back.f_code.co_name
            print(f"Attempting to acquire lock \"{self.name}\" in function \"{self.func_name}\"", flush=True)

        if self.pre_acquire_callback is not None:
            self.pre_acquire_callback()

        self.lock.acquire(blocking=True, timeout=self.timeout)
        if not self.is_locked():
            raise LockAcquisitionError(f"Failed to acquire lock \"{self.name}\" in function \"{self.func_name}\"")

        if _DEBUG_LOCK_PRINT:
            print(f"Acquired lock \"{self.name}\" in function \"{self.func_name}\"", flush=True)
        return self
    
    def __exit__(self, exc_type, exc_value, traceback):
        if self.is_locked():
            self.lock.release()
        if _DEBUG_LOCK_PRINT:
            print(f"Released lock \"{self.name}\" in function \"{self.func_name}\"", flush=True)
        
        if self.post_release_callback is not None:
            self.post_release_callback()


class CriticalException(Exception):
    """An exception that should still be raised and crash the program, even when caught for logging purposes."""
    pass


def try_log_exception_peacefully(
    func: Optional[Callable[P, R]] = None, 
    *, 
    capture_locals: Optional[Iterable[str]] = None
) -> Callable[P, Optional[R]]:
    """Decorate a function to log exceptions.
    
    If an exception is raised, it will be logged and the exception will be ignored, unless the exception is a 
    CriticalException, in which case it will be re-raised. Caught exceptions will return None.

    Can be used either like `@try_log_exception_peacefully` or `@try_log_exception_peacefully(args...)`.

    Args:
        func: Function to decorate. If not provided, will return a decorator that can be used to decorate a function.
        capture_locals: Iterable of local variable names to capture and log.
    """
    # Function was used like `@try_log_exception_peacefully(...)`
    if func is None:
        def decorator(f: Callable[P, R]) -> Callable[P, Optional[R]]:
            return try_log_exception_peacefully(f, capture_locals=capture_locals)
        return decorator

    capture_locals = list(capture_locals) if capture_locals is not None else []
    
    def wrapper(*args: P.args, **kwargs: P.kwargs) -> Optional[R]:
        try:
            return func(*args, **kwargs)
        except Exception as e:
            cl_str = ""

            if len(capture_locals) > 0:
                if e.__traceback__ is not None:
                    frame = e.__traceback__.tb_next.tb_frame
                    cl_str = f"\n\tCaptured locals:\n"
                    local_vars = frame.f_locals

                    for local_var in capture_locals:
                        if local_var in local_vars:
                            s = pprint.pformat(local_vars[local_var], indent=8)
                            cl_str += f"\t{local_var} ({type(local_vars[local_var]).__name__}): {s}\n"
                        else:
                            cl_str += f"\t\t{local_var}: not found\n"

            exc_str = f"Exception in \"{func.__name__}\": \n{traceback.format_exc()}{cl_str}"
            warning(exc_str)

            if isinstance(e, CriticalException):
                raise e
            
            return None
    return wrapper


TimeType = float
"""The type that we use to handle time-related operations."""


def currtime() -> TimeType:
    """Get the current time in seconds.

    Currently (pun intended), this returns time.monotonic(), which returns time since computer power-on on linux.
    """
    return time.monotonic()


def timediff(start_time: TimeType) -> TimeType:
    """Get the time difference in seconds between the current time and the given start time."""
    return currtime() - start_time


def timestamped_class(cls: Type[T]) -> Type[T]:
    """Decorate a class to add a timestamp attribute to the class.
    
    The timestamp is set to the latest time that:
    - The __init__ method is called
    - The __post_init__ method is called (for dataclasses)
    - The object is created using the __new__ method (in case __init__ or __post_init__ are not called)
    """
    old_new = cls.__new__

    def new_new(cls: Type[T], *args: Any, **kwargs: Any) -> T:
        if old_new == object.__new__:
            obj = object.__new__(cls)
        else:
            obj = old_new(cls, *args, **kwargs)

        def _make_timestamped_init_func(cls: Type[T], func_name: str) -> None:
            if not hasattr(cls, func_name) or not callable(getattr(cls, func_name)):
                return
            
            old_func = getattr(cls, func_name)
            def new_func(self: cls, *args: Any, **kwargs: Any) -> None:
                old_func(self, *args, **kwargs)
                self.timestamp = currtime()
            setattr(cls, func_name, new_func)

        # Wrap the class's __init__ and __post_init__ methods, but only do it once
        if not hasattr(cls, '_timestamped_init_wrapped'):
            setattr(cls, '_timestamped_init_wrapped', True)
            _make_timestamped_init_func(cls, '__init__')
            _make_timestamped_init_func(cls, '__post_init__')
        
        # Place a manual timestamp here just in case the object never gets its __init__ or __post_init__ called
        obj.timestamp = currtime()
        
        return obj

    cls.__new__ = new_new
    return cls
    

def exponential_timeout(
    timeout_seconds: float, 
    func: Callable[P, Union[None, R]],
    func_args: P.args = (),
    func_kwargs: P.kwargs = {},
    start_delay: float = 1e-4,  # 100 microseconds
    max_delay: float = 1e-2,  # 10 milliseconds
    mult: float = 1.5, 
) -> Union[R, None]:
    """Perform an exponential timeout.

    Will run the given function, sleeping for longer and longer until the function returns a non-None value, or the
    timeout is reached (in which case a TimeoutError is raised).
    
    Args:
        timeout_seconds: The timeout in seconds.
        func: The function to execute. Should return None if the operation is not complete, and a non-None result
            if it is.
        func_args: Arguments to pass to the function.
        func_kwargs: Keyword arguments to pass to the function.
        start_delay: The starting delay in seconds.
        max_delay: The maximum delay in seconds.
        mult: Delay is multiplied by this factor after each iteration. (but not allowed to exceed max_delay)
    
    Raises:
        TimeoutError: If the function does not return a non-None value before the timeout is reached.
    """
    timeout_tt = TimeType(timeout_seconds)

    init_time = currtime()
    delay = start_delay
    while True:
        result = func(*func_args, **func_kwargs)
        if result is not None:
            return result
        time.sleep(delay)

        if timediff(init_time) > timeout_tt:
            raise TimeoutError(f"Timeout after {timediff(init_time)} seconds")
        delay = min(delay * mult, max_delay, timeout_tt - timediff(init_time))


def listdirfull(dir: str, startswith: Optional[str] = None, endswith: Optional[str] = None) -> List[str]:
    """List all files in the given directory, returning a list of full file paths.
    
    Args:
        dir: The directory to list files from.
        startswith: If provided, only list files that start with this string.
        endswith: If provided, only list files that end with this string.
    """
    return [os.path.join(dir, f) for f in os.listdir(dir) \
        if (startswith is None or f.startswith(startswith)) and (endswith is None or f.endswith(endswith))]


_TO_SNAKE_CASE_RE = re.compile(r'(?<!^)(?=[A-Z])')

def to_snake_case(name: str) -> str:
    """Convert a string to snake case."""
    return _TO_SNAKE_CASE_RE.sub(r'_', name).lower()


def dont_allow_on_fork(func):
    """Decorator to prevent a method from being called after fork."""

    @functools.wraps(func)
    def wrapper(self, *args, **kwargs):
        if hasattr(self, 'is_forked') and self.is_forked:
            warning(
                f"Cannot call method `{self.__class__.__name__}.{func.__name__}` after fork. "
                f"{os.getpid()}, {threading.get_ident()}"
            )
            return
        return func(self, *args, **kwargs)

    return wrapper


class FeatherThread(threading.Thread):
    """A custom threading.Thread. Currently can be used to debug print messages."""
    def __init__(self, target, name=None, args=tuple(), kwargs=None, daemon=True):
        kwargs = {} if kwargs is None else kwargs
        name = name if name is not None else f"FeatherThread-{target.__name__}"
        super().__init__(target=target, name=name, args=args, kwargs=kwargs, daemon=daemon)



"""
Pinocchio model loader for robots.

This module provides PinocchioModel class for loading and managing Pinocchio models
and collision geometry, making them available for multiple tasks (kinematics, dynamics,
collision checking, visualization, etc.).
"""

import logging
import pinocchio as pin
import numpy as np
from pathlib import Path
from feathersdk.utils.feathertypes import List, Optional
from feathersdk.robot.controls.planner import EasyIKPlanner, quaternion_to_rotation_matrix, rotation_matrix_to_quaternion
from typing import Tuple, Union
from feathersdk.utils.logger import info


logger = logging.getLogger(__name__)


class PinocchioModel:
    """
    Reusable Pinocchio model and geometry loader.
    
    This class loads and manages the Pinocchio model and collision geometry,
    making them available for multiple tasks (kinematics, dynamics, collision checking, etc.).
    """
    
    def __init__(
        self, 
        urdf_path: str, 
        package_dir: Optional[str] = None,
        load_collision: bool = True
    ):
        """
        Initialize Pinocchio model from URDF.
        
        Args:
            urdf_path: Path to the URDF file
            package_dir: Directory containing mesh files (for package:// paths)
            load_collision: If True, load collision geometry model
        """
        self.urdf_path = Path(urdf_path)
        
        # Determine package directory
        if package_dir is None:
            # Try to auto-detect: look for package.xml to find the package root
            current = self.urdf_path.parent
            while current != current.parent:  # Stop at filesystem root
                if (current / "package.xml").exists():
                    # Found package root (Nov11_description/)
                    # package_dir should be parent of this
                    self.package_dir = current.parent
                    logger.info(f"Auto-detected package_dir: {self.package_dir}")
                    break
                current = current.parent
            else:
                # Fallback: assume package is 2 levels up from URDF
                self.package_dir = self.urdf_path.parent.parent.parent
                logger.info(f"Using inferred package_dir: {self.package_dir}")
        else:
            self.package_dir = Path(package_dir)
        
        # Load the kinematic model
        self.model = pin.buildModelFromUrdf(str(self.urdf_path))
        self.data = self.model.createData()
        self.last_q = pin.neutral(self.model)
        self.ik_planner = EasyIKPlanner(self.model)
        
        # Load collision geometry if requested
        self.geom_model = None
        self.geom_data = None
        if load_collision:
            self._load_collision_geometry()
        
        logger.info(f"✓ Model loaded: {self.model.name}")
        self.n_frames = getattr(self.model, 'nframes', len(self.model.names))
        self.link_names = [name for name in self.model.names]
        logger.info(f"✓ Number of frames: {self.n_frames}")
        logger.info(f"✓ Number of joints: {self.model.njoints}")
        if self.geom_model:
            logger.info(f"✓ Number of collision geometries: {len(self.geom_model.geometryObjects)}")

    def update_q(self, q: np.ndarray):
        self.last_q = q
        self.ik_planner.update_q(q)

    def get_joint_pose(self, joint_id: Union[int, str], is_dict: bool = False) -> np.ndarray:
        """Get the pose of the given joint as a pose."""
        if isinstance(joint_id, str):
            joint_id = self.model.getJointId(joint_id)
        if is_dict:
            return {
                "position": self.data.oMi[joint_id].translation,
                "rotation": self.data.oMi[joint_id].rotation
            }
        else:
            return np.concatenate([self.data.oMi[joint_id].translation, rotation_matrix_to_quaternion(self.data.oMi[joint_id].rotation)])
    
    def _load_collision_geometry(self):
        """Load collision geometry from URDF."""
        logger.info(f"Loading collision geometry with package_dirs: [{self.package_dir}]")
        
        try:
            self.geom_model = pin.buildGeomFromUrdf(
                self.model,
                str(self.urdf_path),
                pin.GeometryType.COLLISION,
                None,  # geometry_model=None
                [str(self.package_dir)]  # package_dirs as list
            )
            # 2. Add all possible collision pairs (Self-Collision)
            # This includes every link against every other link.
            self.geom_model.addAllCollisionPairs()
            # 3. Optimization: Remove adjacent links 
            # (Links connected by a joint usually shouldn't collide with each other)
            # pin.removeCollisionPairs(self.model, self.geom_model, str(self.urdf_path))
            self.geom_data = self.geom_model.createData()
            self.collision_data = pin.GeometryData(self.geom_model)

            # Needed to remove collisions of neighboring links and when the torso is at the bottom.
            self.check_collision({
                "EpL": -1.57,
                "EpR": 1.57,
                "TvC": -0.55
            })
            self.remove_all_current_collisions()
            logger.info(f"✓ Successfully loaded collision geometry")
        except Exception as e:
            # Fallback: try without package_dirs
            logger.warning(f"⚠ Loading with package_dirs failed, trying without...")
            try:
                self.geom_model = pin.buildGeomFromUrdf(
                    self.model,
                    str(self.urdf_path),
                    pin.GeometryType.COLLISION
                )
                self.geom_data = self.geom_model.createData()
                self.geom_model.addAllCollisionPairs()
                # 3. Optimization: Remove adjacent links 
                # (Links connected by a joint usually shouldn't collide with each other)
                # pin.removeCollisionPairs(self.model, self.geom_model, "/home/hoa/Workspace/featheros/jetson_kernel/on_filesys/feathersys/user_config/conf/simple_URDF_description/meshes/robot.srdf")
                
                self.geom_data = self.geom_model.createData()
                self.collision_data = pin.GeometryData(self.geom_model)

                # Needed to remove collisions of neighboring links and when the torso is at the bottom.
                self.check_collision({
                    "EpL": -1.57,
                    "EpR": 1.57,
                    "TvC": -0.55
                })
                self.remove_all_current_collisions()

                logger.info(f"✓ Successfully loaded collision geometry (without package_dirs)")
            except Exception as e2:
                logger.error(f"\n❌ Failed to load collision geometry")
                logger.error(f"   package_dirs tried: [{self.package_dir}]")
                logger.error(f"   Errors: {e}, {e2}")
                raise RuntimeError(
                    f"Failed to load collision geometry. "
                    f"package_dirs: [{self.package_dir}]. "
                    f"Last error: {e2}"
                )
        
        # Initialize geometry data with neutral configuration
        # try:
        #     q_neutral = pin.neutral(self.model)
        #     pin.forwardKinematics(self.model, self.data, q_neutral)
        #     pin.updateGeometryPlacements(
        #         self.model, self.data, self.geom_model, self.geom_data, q_neutral
        #     )
        #     logger.info("✓ Geometry data initialized successfully")
        # except Exception as e:
        #     logger.warning(f"⚠ Warning: Could not initialize geometry data: {e}")
    
    def check_collision_path(self, q_start: np.ndarray, q_end: np.ndarray) -> bool:
        # Implement me
        pass

    def check_collision(self, q_updates: Union[np.ndarray, dict], stop_at_first_collision: bool = False) -> bool:
        q = self.q_from_q_updates(q_updates)
        return pin.computeCollisions(
            self.model, 
            self.data, 
            self.geom_model, 
            self.geom_data, 
            q,
            stop_at_first_collision
        )

    def get_collision_pairs_from_last_collision_check(self) -> List[Tuple[str, str]]:
        # Find which pairs are actually in collision
        collisions = []
        for i, result in enumerate(self.geom_data.collisionResults):
            if result.isCollision():
                # Get the pair definition
                pair = self.geom_model.collisionPairs[i]
                
                # Get the names of the geometry objects
                name1 = self.geom_model.geometryObjects[pair.first].name
                name2 = self.geom_model.geometryObjects[pair.second].name
                
                collisions.append((name1, name2))
        return collisions
    
    def remove_all_current_collisions(self):
        collisions = self.get_collision_pairs_from_last_collision_check()
        for link_a, link_b in collisions:
            # 1. Get the IDs of the collision objects
            # Note: These names must match the names in your geometry_model
            if self.geom_model.existGeometryName(link_a) and self.geom_model.existGeometryName(link_b):
                id_a = self.geom_model.getGeometryId(link_a)
                id_b = self.geom_model.getGeometryId(link_b)
                
                # 2. Create a CollisionPair object
                pair = pin.CollisionPair(id_a, id_b)
                
                # 3. Remove the pair from the model
                if self.geom_model.existCollisionPair(pair):
                    self.geom_model.removeCollisionPair(pair)
                    info(f"Removed: {link_a} <-> {link_b}")
            else:
                info(f"Warning: One of these links was not found: {link_a}, {link_b}")

        # 4. Critical: Re-generate the geometry data to apply changes
        self.geom_data = pin.GeometryData(self.geom_model)

    def q_from_q_updates(self, q_updates: Union[np.ndarray, dict]) -> np.ndarray:
        q = q_updates
        if isinstance(q, dict):
            q = self.last_q
        for name in q_updates.keys():
            q[self.model.joints[self.model.getJointId(name)].idx_q] = q_updates[name]
        if q.ndim > 1:
            q = q.flatten()
        q = np.asarray(q, dtype=np.float64)
        return q

    def update_kinematics(self, q_updates: Union[np.ndarray, dict]):
        """
        Update forward kinematics for the given configuration.
        
        Args:
            q: Joint configuration vector (nq x 1) or a map of the joint names to the joint values
        """
        q = self.q_from_q_updates(q_updates)
        self.update_q(q)
        
        if len(q) != self.model.nq:
            raise ValueError(
                f"Configuration size mismatch: got {len(q)}, expected {self.model.nq}"
            )
        
        pin.forwardKinematics(self.model, self.data, q)
    
    def update_geometry(self, q: np.ndarray):
        """
        Update geometry placements based on joint configuration.
        
        Args:
            q: Joint configuration vector (nq x 1)
        """
        if self.geom_model is None:
            raise RuntimeError("Collision geometry not loaded. Set load_collision=True in __init__")
        
        self.update_kinematics(q)
        pin.updateGeometryPlacements(
            self.model, self.data, self.geom_model, self.geom_data, q
        )
    
    def get_link_names(self) -> List[str]:
        return self.link_names

    def get_geometry_names(self) -> List[str]:
        """Get list of all collision geometry names."""
        if self.geom_model is None:
            return []
        return [geom.name for geom in self.geom_model.geometryObjects]
